/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    MemoryX::ArmarXObjects::GraphImportExport
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <VirtualRobot/VirtualRobot.h>

#include "GraphImportExport.h"

#include <armarx/navigation/location/aron/Location.aron.generated.h>
#include <armarx/navigation/location/constants.h>

#include <armarx/navigation/graph/aron/Graph.aron.generated.h>
#include <armarx/navigation/graph/constants.h>
#include <armarx/navigation/graph/Graph.h>
#include <armarx/navigation/graph/Visu.h>


#include <MemoryX/libraries/memorytypes/MemoryXTypesObjectFactories.h>
#include <MemoryX/core/MemoryXCoreObjectFactories.h>

#include <RobotAPI/libraries/core/FramedPose.h>
#include <RobotAPI/libraries/armem/core/Commit.h>
#include <RobotAPI/libraries/armem/core/aron_conversions.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <iomanip>


namespace armarx::nav
{

    GraphImportExport::GraphImportExport()
    {
        properties.memoryName = "Navigation";
        properties.locationCoreSegmentName = loc::coreSegmentName;
        properties.graphCoreSegmentName = graph::coreSegmentName;
    }


    armarx::PropertyDefinitionsPtr GraphImportExport::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def = new ComponentPropertyDefinitions(getConfigIdentifier());

        def->component(proxies.priorKnowledge);
        def->component(proxies.graphNodePoseResolver, "GraphNodePoseResolver");

        return def;
    }


    void GraphImportExport::onInitComponent()
    {
        // Topics and properties defined above are automagically registered.

        // Keep debug observer data until calling `sendDebugObserverBatch()`.
        setDebugObserverBatchModeEnabled(true);
    }


    void GraphImportExport::onConnectComponent()
    {
        // Get proxies.
        if (proxies.priorKnowledge->hasGraphSegment())
        {
            proxies.graphSegment = proxies.priorKnowledge->getGraphSegment();
        }
        proxies.navigationWriter = memoryNameSystem.useWriter(armem::MemoryID(properties.memoryName));

        // Setup and start the remote GUI.
        refreshScenes();
        RemoteGui_startRunningTask();
    }


    void GraphImportExport::onDisconnectComponent()
    {
    }


    void GraphImportExport::onExitComponent()
    {
    }


    std::string GraphImportExport::getDefaultName() const
    {
        return "GraphImportExport";
    }


    void GraphImportExport::createRemoteGuiTab(const std::vector<std::string>& sceneNames)
    {
        using namespace armarx::RemoteGui::Client;

        tab.sceneComboBox.setOptions(sceneNames);
        tab.sceneRefreshButton.setLabel("Refresh");

        tab.providerSegmentLine.setValue(getName());

        tab.dryRun.setValue(false);
        tab.visuEnabled.setValue(true);

        tab.locationsMemoryxToArmemButton.setLabel("Locations MemoryX -> ArMem");
        tab.locationsArmemToMemoryxButton.setLabel("Locations ArMem -> MemoryX (WIP)");
        tab.locationsClearArMemButton.setLabel("Clear ArMem Locations");

        tab.graphMemoryxToArmemButton.setLabel("Graph MemoryX -> ArMem");
        tab.graphArmemToMemoryxButton.setLabel("Graph ArMem -> MemoryX (WIP)");
        tab.graphClearArMemButton.setLabel("Clear ArMem Graphs");


        GridLayout grid;
        int row = 0;
        {
            grid.add(Label("Scene:"), {row, 0}).add(tab.sceneComboBox, {row, 1}).add(tab.sceneRefreshButton, {row, 2});
            ++row;

            grid.add(Label("Provider Segment:"), {row, 0}).add(tab.providerSegmentLine, {row, 1});
            ++row;

            grid.add(Label("Dry Run:"), {row, 0}).add(tab.dryRun, {row, 1});
            ++row;

            grid.add(Label("Enable visu:"), {row, 0}).add(tab.visuEnabled, {row, 1});
            ++row;

            grid.add(tab.locationsMemoryxToArmemButton, {row, 0}).add(tab.locationsArmemToMemoryxButton, {row, 1})
                    .add(tab.locationsClearArMemButton, {row, 2});
            ++row;

            grid.add(tab.graphMemoryxToArmemButton, {row, 0}).add(tab.graphArmemToMemoryxButton, {row, 1})
                    .add(tab.graphClearArMemButton, {row, 2});
            ++row;
        }

        VBoxLayout root = {grid, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }


    void GraphImportExport::RemoteGui_update()
    {
        if (tab.sceneRefreshButton.wasClicked())
        {
            refreshScenes();
        }

        if (tab.locationsMemoryxToArmemButton.wasClicked())
        {
            locationsMemoryxToArmem(tab.sceneComboBox.getValue());
        }
        if (tab.locationsArmemToMemoryxButton.wasClicked())
        {
            locationsArmemToMemoryx(tab.sceneComboBox.getValue());
        }
        if (tab.graphMemoryxToArmemButton.wasClicked())
        {
            graphMemoryxToArmem(tab.sceneComboBox.getValue());
        }
        if (tab.graphArmemToMemoryxButton.wasClicked())
        {
            graphArmemToMemoryx(tab.sceneComboBox.getValue());
        }

        if (tab.locationsClearArMemButton.wasClicked())
        {
            clearArMemProviderSegment(getLocationProviderSegmentID());
        }
        if (tab.graphClearArMemButton.wasClicked())
        {
            clearArMemProviderSegment(getGraphProviderSegmentID());
        }
    }


    void GraphImportExport::refreshScenes()
    {
        const ::Ice::StringSeq sceneNames = proxies.graphSegment->getScenes();
        createRemoteGuiTab(sceneNames);
    }


    void GraphImportExport::clearArMemProviderSegment(const armem::MemoryID& providerSegmentID)
    {
        const bool clearWhenExists = true;
        auto result = proxies.navigationWriter.addSegment(providerSegmentID, clearWhenExists);
        if (result.success)
        {
            ARMARX_IMPORTANT << "Cleared ArMem provider segment " << providerSegmentID << ".";
        }
        else
        {
            ARMARX_WARNING << result.errorMessage;
        }
    }


    void GraphImportExport::locationsMemoryxToArmem(const std::string& sceneName)
    {
        const armem::Time time = armem::Time::now();
        armem::Commit commit;

        memoryx::GraphNodeBaseList graphNodes = proxies.graphSegment->getNodesByScene(sceneName);
        for (memoryx::GraphNodeBasePtr& node : graphNodes)
        {
            ARMARX_CHECK_NOT_NULL(node);
            if (not node->isMetaEntity())
            {
                // ID is just some random MongoDB hash
                const std::string nodeId = node->getId();
                // This is the readable name entered in the GUI.
                const std::string name = node->getName();

                armarx::FramedPosePtr pose = armarx::FramedPosePtr::dynamicCast(node->getPose());
                ARMARX_CHECK_NOT_NULL(pose);

                FramedPosePtr globalNodePose = FramedPosePtr::dynamicCast(proxies.graphNodePoseResolver->resolveToGlobalPose(node));
                ARMARX_CHECK_NOT_NULL(globalNodePose);

                // `pose` and `globalNodePose` seem to be identical. Is the last step necessary?
                // Maybe `pose` could be non-global.

                ARMARX_VERBOSE
                        << std::setprecision(2) << std::fixed
                        << "Processing node " << (commit.updates.size() + 1)
                        << "\n- ID: \t" << nodeId
                        << "\n- Name: \t" << name
                        << "\n- Pose: \n" << pose->toEigen()
                        << "\n- Resolved global pose: \n" << globalNodePose->toEigen()
                           ;

                nav::loc::arondto::Location data;
                data.globalRobotPose = globalNodePose->toEigen();

                armem::EntityUpdate& update = commit.add();
                update.entityID = getLocationProviderSegmentID().withEntityName(name);
                update.timeCreated = time;
                update.instancesData = { data.toAron() };
            }
        }

        if (not tab.dryRun.getValue())
        {
            armem::CommitResult result = proxies.navigationWriter.commit(commit);
            if (result.allSuccess())
            {
                ARMARX_IMPORTANT << "Successfully exported " << result.results.size() << " locations from MemoryX to ArMem.";
            }
            else
            {
                ARMARX_WARNING << result.allErrorMessages();
            }
        }
        else
        {
            ARMARX_VERBOSE << "Dry Run - skipping commit.";
        }
    }


    void GraphImportExport::locationsArmemToMemoryx(const std::string& sceneName)
    {
        ARMARX_IMPORTANT << "locationsArmemToMemoryx() is WIP!";
        (void) sceneName;
    }


    void GraphImportExport::graphMemoryxToArmem(const std::string& sceneName)
    {
        memoryx::GraphNodeBaseList graphNodes = proxies.graphSegment->getNodesByScene(sceneName);
        nav::graph::Graph graph = toArmemGraph(graphNodes);

        // Build ARON Graph
        nav::graph::arondto::Graph aron;
        toAron(aron, graph);


        if (tab.visuEnabled.getValue())
        {
            nav::graph::GraphVisu visu;
            viz::Layer layer = arviz.layer("Graph '" + sceneName + "'");
            visu.draw(layer, graph);

            ARMARX_VERBOSE << "Visualize graph '" << sceneName << "' ...";
            arviz.commit(layer);
        }


        // Build commit
        const armem::Time time = armem::Time::now();
        armem::EntityUpdate update;
        update.entityID = getGraphProviderSegmentID().withEntityName(sceneName);
        update.timeCreated = time;
        update.instancesData = { aron.toAron() };

        if (not tab.dryRun.getValue())
        {
            armem::EntityUpdateResult result = proxies.navigationWriter.commit(update);
            if (result.success)
            {
                ARMARX_IMPORTANT << "Successfully exported graph '" << sceneName << "' from MemoryX to ArMem.";
            }
            else
            {
                ARMARX_WARNING << result.errorMessage;
            }
        }
        else
        {
            ARMARX_VERBOSE << "Dry Run - skipping commit.";
        }
    }


    void GraphImportExport::graphArmemToMemoryx(const std::string& sceneName)
    {
        ARMARX_IMPORTANT << "graphArmemToMemoryx() is WIP!";
        (void) sceneName;
    }


    armem::MemoryID GraphImportExport::getLocationProviderSegmentID()
    {
        return armem::MemoryID(properties.memoryName,
                               properties.locationCoreSegmentName,
                               tab.providerSegmentLine.getValue());
    }

    armem::MemoryID GraphImportExport::getGraphProviderSegmentID()
    {
        return armem::MemoryID(properties.memoryName,
                               properties.graphCoreSegmentName,
                               tab.providerSegmentLine.getValue());
    }

    nav::graph::Graph GraphImportExport::toArmemGraph(const memoryx::GraphNodeBaseList& graphNodes)
    {
        nav::graph::Graph graph;
        std::map<std::string, nav::graph::Graph::Vertex> vertexMap;

        // Add nodes
        semrel::ShapeID nextVertexID { 0 };
        for (memoryx::GraphNodeBasePtr node : graphNodes)
        {
            ARMARX_CHECK_NOT_NULL(node);
            if (not node->isMetaEntity())
            {
                // This is the readable name entered in the GUI.
                const std::string name = node->getName();

                FramedPosePtr globalNodePose = FramedPosePtr::dynamicCast(proxies.graphNodePoseResolver->resolveToGlobalPose(node));
                ARMARX_CHECK_NOT_NULL(globalNodePose);

                ARMARX_VERBOSE << "\n- Adding node: \t" << name;

                nav::graph::Graph::Vertex& vertex = vertexMap.emplace(name, graph.addVertex(nextVertexID)).first->second;
                vertex.attrib().aron.vertexID = static_cast<long>(nextVertexID);
                toAron(vertex.attrib().aron.locationID, getLocationProviderSegmentID().withEntityName(name));
                vertex.attrib().aron.globalRobotPose = globalNodePose->toEigen();

                nextVertexID++;
            }
        }

        // Add edges
        for (memoryx::GraphNodeBasePtr node : graphNodes)
        {
            const auto& sourceVertex = vertexMap.at(node->getName());
            for (int i = 0; i < node->getOutdegree(); i++)
            {
                auto adjacent = memoryx::GraphNodeBasePtr::dynamicCast(node->getAdjacentNode(i)->getEntity());
                ARMARX_CHECK_NOT_NULL(adjacent);
                const auto& targetVertex = vertexMap.at(adjacent->getName());

                ARMARX_VERBOSE << "\n- Adding edge: \t" << node->getName() << " \t-> " << adjacent->getName();
                nav::graph::Graph::Edge edge = graph.addEdge(sourceVertex, targetVertex);
                edge.attrib().aron.sourceVertexID = static_cast<long>(sourceVertex.attrib().aron.vertexID);
                edge.attrib().aron.targetVertexID = static_cast<long>(targetVertex.attrib().aron.vertexID);
            }
        }

        return graph;
    }

}
